#include "hashtab.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define ALLOC_SIZE 10

unsigned Hash(unsigned char *ptr,int len,unsigned top)
{
  int x=0;
  unsigned total=0,g=0;
  
  while (x<len)
  {
    total=(total<<4)+ptr[x++];
    if ((g=(total&0xf0000000))!=0)
    {
      total=total^(g>>24);
      total=total^g;
    }
  }
  total^=(total/top);
  return total%top;
}

HashTable HTCreate(int buckets,int key_size,int rec_size,
                   HTHashProc hash_proc,
                   HTCopyProc copy_proc,
                   HTFreeProc free_proc,
                   HTCompareProc compare_proc)
{
  HashTable ht;
  int x;
  
  if ((ht.data=(HashTableEntry*)malloc(buckets*sizeof(HashTableEntry)))==NULL)
  {
    ht.buckets=0;
    ht.key_size=0;
    ht.rec_size=0;
  }
  else
  {
    ht.buckets=buckets;
    ht.key_size=key_size;
    ht.rec_size=rec_size;
    for (x=0;x<buckets;x++)
    {
      ht.data[x].entries=0;
      ht.data[x].allocated=0;
      ht.data[x].data=NULL;
    }
  }
  if (hash_proc==NULL)
    ht.hash_proc=(HTHashProc)Hash;
  else
    ht.hash_proc=hash_proc;
  if (copy_proc==NULL)
    ht.copy_proc=(HTCopyProc)memcpy;
  else
    ht.copy_proc=copy_proc;
  if (free_proc==NULL)
    ht.free_proc=(HTFreeProc)free;
  else
    ht.free_proc=free_proc;
  if (compare_proc==NULL)
    ht.compare_proc=(HTCompareProc)memcmp;
  else
    ht.compare_proc=compare_proc;
  return ht;
}

void HTAddRecord(HashTable ht,void *record)
{
  int x;
  
  x=ht.hash_proc((unsigned char*)record,ht.key_size,ht.buckets);
  if (ht.data[x].entries+1>ht.data[x].allocated)
  {
    ht.data[x].allocated+=ALLOC_SIZE;
    if ((ht.data[x].data=(void**)realloc((void*)ht.data[x].data,(ht.data[x].allocated)*sizeof(void*)))==NULL)
      return;
  }
  ht.data[x].data[ht.data[x].entries]=malloc(ht.rec_size);
  ht.copy_proc(ht.data[x].data[ht.data[x].entries],record,ht.rec_size);
  ht.data[x].entries++;
} 

int HTLookup(HashTable ht,void *key)
{
  int x,y;
  
  x=ht.hash_proc((unsigned char*)key,ht.key_size,ht.buckets);
  for (y=0;y<ht.data[x].entries;y++)
  {
    if (ht.compare_proc(ht.data[x].data[y],key,ht.key_size)==0)
      return 1;
  }
  return 0;
}

int HTGetRecord(HashTable ht,void *key,void *record)
{
  int x,y;
  
  x=ht.hash_proc((unsigned char*)key,ht.key_size,ht.buckets);
  for (y=0;y<ht.data[x].entries;y++)
  {
    if (ht.compare_proc(ht.data[x].data[y],key,ht.key_size)==0)
    {
      ht.copy_proc(record,ht.data[x].data[y],ht.rec_size);
      return 1;
    }
  }
  return 0;
}

int HTDeleteRecord(HashTable ht,void *key)
{
  int x,y;

  x=ht.hash_proc((unsigned char*)key,ht.key_size,ht.buckets);
  for (y=0;y<ht.data[x].entries;y++)
  {
    if (ht.compare_proc(ht.data[x].data[y],key,ht.key_size)==0)
    {
      ht.free_proc(ht.data[x].data[y]);
      ht.data[x].entries--;
      if (ht.data[x].entries>y)
      {
        memmove(&ht.data[x].data[y],&ht.data[x].data[y+1],(ht.data[x].entries-y)*sizeof(void*));
      }
      if (ht.data[x].entries<=ht.data[x].allocated-ALLOC_SIZE)
      {
        ht.data[x].allocated-=ALLOC_SIZE;
        ht.data[x].data=(void**)realloc(ht.data[x].data,ht.data[x].allocated*sizeof(void*));
      }
      return 1;
    }
  }
  return 0;
}

void HTDestroy(HashTable ht)
{
  int x,y;
  
  for (x=0;x<ht.buckets;x++)
  {
    for (y=0;y<ht.data[x].entries;y++)
      ht.free_proc(ht.data[x].data[y]);
    if (ht.data[x].data && ht.data[x].entries)
      free(ht.data[x].data);
  }
  free(ht.data);
}

void HTForEach(HashTable ht,HTForEachProc feproc)
{
  int x,y;
  
  for (x=0;x<ht.buckets;x++)
  {
    for (y=0;y<ht.data[x].entries;y++)
      feproc(ht.data[x].data[y]);
  }
}

int HTNumberOfRecords(HashTable ht)
{
  int tot=0,x;
  
  for (x=0;x<ht.buckets;x++)
  {
    tot+=ht.data[x].entries;
  }
  return tot;
}  
