#include "jbsql.h"

HINSTANCE hInst,g_hInst;
HWND      g_hWnd;
TCHAR     g_szClassName[] = TEXT("VListVwClass");
HWND hwndListView;
HWND hwndTextBox;
HWND hwndSplitter;
HMENU g_menu;
TextWidget *g_text=NULL;
char executable[MAX_PATH];
HashTable vht;
DWORD g_tid;
HANDLE pthandle;

#define SCREENBUFSIZE 1000

char *myvsprintf(const char *fmt, va_list vl)
{
  char *str=NULL;
  int size=0;

  size=SCREENBUFSIZE;
  str=(char*)malloc(size);
  for (;;)
  {
    if (_vsnprintf(str,size-1,fmt,vl)==-1)
    {
      size+=SCREENBUFSIZE;
      if ((str=(char*)realloc(str,size))==NULL)
      {
        MessageBox(NULL,"myvsprintf(): Out of memory",NULL,MB_OK|MB_ICONERROR);
        break;
      }
    }
    else
      break;
  }
  return str;
}

void Log(char *fmt,...)
{
  va_list param;
  char *str;
  FILE *f;

  if (g_text && g_text->options[DEBUG].value.int_value)
  {
    va_start(param,fmt);
    str=myvsprintf(fmt,param);
    va_end(param);
    f=fopen("jbsql.log","a+");
    if (f==NULL)
		  return;
    if (str[strlen(str)-1]=='\n')
      fprintf(f,"%s",str);
    else
      fprintf(f,"%s\n",str);
    fclose(f);
    free(str);
  }
}

void Error(HWND hWnd,char *fmt,...)
{
  va_list param;
  char *str;

  va_start(param,fmt);
  str=myvsprintf(fmt,param);
  va_end(param);
  MessageBox(hWnd,str,NULL,MB_OK|MB_ICONERROR);
  free(str);
}

void DisplayLastError(HWND hWnd)
{
  LPVOID lpMsgBuf;
  
  FormatMessage( 
    FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,
    NULL,
    GetLastError(),
    MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
    (LPTSTR) &lpMsgBuf,    0,    NULL );
  Error( hWnd, lpMsgBuf);
  LocalFree( lpMsgBuf );
}

void SetWindowTitle(char *username,char *database)
{
  char title[200];

  if (username[0]!='\0')
  {
    if (database[0]!='\0')
      sprintf(title,"JB/SQL - %s@%s",username,database);
    else
      sprintf(title,"JB/SQL - %s@local",username);
  }
  else
    sprintf(title,"JB/SQL - Disconnected");
  SetWindowText(g_hWnd,title);
}

LRESULT CALLBACK LoginProc( HWND hWnd,
                            UINT uMessage,
                            WPARAM wParam,
                            LPARAM lParam)
{
  static TextWidget *text;

	switch (uMessage)
	{
    case WM_INITDIALOG:
      {
        int x; 

        text=(TextWidget *)lParam;
        for (x=0;x<10;x++)
        {
          if (last_usernames[x][0])
          {
            SendDlgItemMessage(hWnd,IDC_USERNAME,CB_ADDSTRING,0,(LPARAM)last_usernames[x]);
            if (x==0)
              SetDlgItemText(hWnd,IDC_USERNAME,last_usernames[x]);
          }
        }
        for (x=0;x<10;x++)
        {
          if (last_databases[x][0])
          {
            SendDlgItemMessage(hWnd,IDC_DATABASE,CB_ADDSTRING,0,(LPARAM)last_databases[x]);
            if (x==0)
              SetDlgItemText(hWnd,IDC_DATABASE,last_databases[x]);
          }
        }
        SetFocus(hWnd);
      }
      SendDlgItemMessage(hWnd,IDC_CONNECT_AS,CB_ADDSTRING,0,(LPARAM)"NORMAL");
      SendDlgItemMessage(hWnd,IDC_CONNECT_AS,CB_ADDSTRING,0,(LPARAM)"SYSOPER");
      SendDlgItemMessage(hWnd,IDC_CONNECT_AS,CB_ADDSTRING,0,(LPARAM)"SYSDBA");
      SendDlgItemMessage(hWnd,IDC_CONNECT_AS,CB_SETCURSEL,0,0);
      break;
		case WM_COMMAND:
			switch (GET_WM_COMMAND_ID(wParam, lParam))
			{
				case ID_LOGON_OK:
				{
					char username[100],password[100],database[100],connect_as[10];

					GetDlgItemText(hWnd,IDC_USERNAME,username,100);
				  GetDlgItemText(hWnd,IDC_PASSWORD,password,100);
				  GetDlgItemText(hWnd,IDC_DATABASE,database,100);
				  GetDlgItemText(hWnd,IDC_CONNECT_AS,connect_as,10);
					if (ConnectToOracle(username,password,database,connect_as)==0)
  					EndDialog(hWnd,TRUE);
          if (ora_connected)
          {
            ClearCommand(g_text,0,0);
            GetDefaultOptions(g_text);
            SetPWD(g_text);
          }
					return TRUE;
				}
				case ID_LOGON_CANCEL:
					EndDialog(hWnd,FALSE);
					return TRUE;
			}
			break;

    case WM_CLOSE:
			EndDialog(hWnd,FALSE);
      return TRUE;

		default:
			break;
	}
  return FALSE;
}

int ShowLoginBox(HWND parenthWnd,TextWidget *text)
{
	FARPROC aboutproc;
	LRESULT res;

	aboutproc=MakeProcInstance((FARPROC)LoginProc,hInst);
	res=DialogBoxParam(hInst,MAKEINTRESOURCE(IDD_LOGON),parenthWnd,(DLGPROC)aboutproc,(LPARAM)text);
	FreeProcInstance(aboutproc);
	if (!res)
		return 0;
	return 1;
}

void NewWindow(TextWidget *text)
{
  char path[1024]="",original_jbsql[MAX_PATH];
  char *pos,*lastpos;
  HANDLE f=INVALID_HANDLE_VALUE;
	PROCESS_INFORMATION p;
  STARTUPINFO si;

  if (strlen(executable)<4 || _stricmp(&executable[strlen(executable)-4],".exe"))
    strcat(executable,".exe");

  if (!strchr(executable,'\\'))
  {
    strcpy(original_jbsql,executable);
    if ((f=CreateFile(executable,
                      0,
                      FILE_SHARE_READ,
                      NULL,
                      OPEN_EXISTING,
                      0,
                      NULL))==INVALID_HANDLE_VALUE)
    {
      GetEnvironmentVariable("PATH",path,1024);
      lastpos=path;
      pos=strchr(lastpos,';');
      while (pos)
      {
        *pos='\0';
        sprintf(original_jbsql,"%s\\%s",lastpos,executable);
        if ((f=CreateFile(original_jbsql,
                          0,
                          FILE_SHARE_READ,
                          NULL,
                          OPEN_EXISTING,
                          0,
                          NULL))!=INVALID_HANDLE_VALUE)
          break;
        lastpos=pos+1;
        pos=strchr(lastpos,';');
      }
      if (f==INVALID_HANDLE_VALUE)
      {
        sprintf(original_jbsql,"%s\\%s",lastpos,executable);
        f=CreateFile(original_jbsql,
                     0,
                     FILE_SHARE_READ,
                     NULL,
                     OPEN_EXISTING,
                     0,
                     NULL);
      }
    }
  }
  else
  {
    strcpy(original_jbsql,executable);
    f=CreateFile(original_jbsql,
                 0,
                 FILE_SHARE_READ,
                 NULL,
                 OPEN_EXISTING,
                 0,
                 NULL);
  }
  if (f==INVALID_HANDLE_VALUE)
  {
    Error(NULL,"Can't find jbql.exe!");
    return;
  }
  CloseHandle(f);
  SaveCommandHistory(text);
  strcat(original_jbsql," \"");
  strcat(original_jbsql,connectstring);
  strcat(original_jbsql,"\"");
	si.cb=sizeof(STARTUPINFO);
	si.lpReserved=NULL;
	si.lpDesktop=NULL;
	si.lpTitle=NULL;
	si.dwFlags=0;
	si.cbReserved2=0;
	si.lpReserved2=NULL;
  if (CreateProcess(NULL,original_jbsql,NULL,NULL,FALSE,0,NULL,NULL,&si,&p)==0)
  {
    DisplayLastError(NULL);
    return;
  }
}

int IsChildWindow(HWND parent,HWND child)
{
  while (child)
  {
    if (child==parent)
      return 1;
    child=GetParent(child);
  }
  return 0;
}

LRESULT CALLBACK MainWndProc( HWND hWnd,
                              UINT uMessage,
                              WPARAM wParam,
                              LPARAM lParam)
{
	switch (uMessage)
	{
		case WM_CREATE:
      {
        TextWidget *text;

        hwndSplitter = CreateSplitter(g_hInst, hWnd);
			  hwndListView = CreateListView(g_hInst, hwndSplitter);
			  InitListView(hwndListView);
			  hwndTextBox  = CreateTextWidget(g_hInst,hWnd,&text);
        text->toolbar=BuildToolbar(text,g_hInst,hWnd);
        g_hWnd=hWnd;
        g_text=text;
			  SetForegroundWindow(hwndTextBox);
        EnableWindow(hwndTextBox,TRUE);
        WPrintf("JB/SQL version %s\n",JBSQL_VERSION);
      }
			break;

    case WM_ACTIVATE:
      SetFocus(hwndTextBox);
      break;

		case WM_SIZE:
      {
        TextWidget *text;

        text=(TextWidget*)GetWindowLong(hwndTextBox, 0);
        SendMessage(text->toolbar,TB_AUTOSIZE,0,0);
			  ResizeSplitter(hwndSplitter,hWnd);
			  break;
      }
		case WM_COMMAND:
      {
        TextWidget *text;

        text=(TextWidget*)GetWindowLong(hwndTextBox, 0);
			  switch (LOWORD(wParam))
			  {
				  case ID_FILE_EXIT:
            if (tid==0)
  					  DestroyWindow(hWnd);
					  break;
          case ID_FILE_CONNECT:
            if (tid==0)
              ShowLoginBox(g_hWnd,text);
            break;
          case ID_NEW_WINDOW:
            NewWindow(text);
            break;
          case ID_FILE_DISCONNECT:
            if (tid==0)
              DisconnectFromOracle();
            break;
          case ID_QUERY_CANCEL:
            AbortQuery();
            break;
          case ID_STATEMENT_PREVIOUS:
            if (tid==0)
              PreviousStatement(text);
            break;
          case ID_STATEMENT_NEXT:
            if (tid==0)
              NextStatement(text);
            break;
          case ID_COMMIT:
            if (tid==0)
            {
              WPrintf("\r");
              ClearCommand(text,0,0);
              InsertLineBuffer(text,"commit;\n");
            }
            break;
          case ID_ROLLBACK:
            if (tid==0)
            {
              WPrintf("\r");
              ClearCommand(text,0,0);
              InsertLineBuffer(text,"rollback;\n");
            }
            break;
          case ID_REPEAT:
            if (tid==0)
              InsertLineBuffer(text,"/\n");
            break;
          case ID_OPTIONS:
            {
              FARPROC dlgproc;
  
              dlgproc=MakeProcInstance((FARPROC)OptionProc,g_hInst);
	            if (DialogBoxParam(g_hInst,MAKEINTRESOURCE(IDD_OPTIONS),hWnd,(DLGPROC)dlgproc,(LPARAM)text)==-1)
                DisplayLastError(hWnd);
	            FreeProcInstance(dlgproc);
            }
            break;
          case ID_SQL_ANALYZER:
            if (text->analyze==NULL)
              text->analyze=ShowSqlAnalyzer(NULL);
            break;
          case ID_SCHEMA_BROWSER:
            if (text->schema==NULL)
              text->schema=ShowSchemaBrowser(NULL);
            break;
          case ID_EXPORT:
            ExportData(text);
            break;
          case ID_COPY:
            if (GetFocus()==hwndTextBox)
            {
              if (text->selection)
                CopyToClipboard(text); 
            }
            else if (GetFocus()==hwndListView)
              CopyFromListView(hwndListView,1);
            break;
          case ID_PASTE:
            if (ptid==0)
            {
   			      pthandle=CreateThread(NULL,0,(LPTHREAD_START_ROUTINE)PasteFromClipboard,text,0,&ptid);
              SetTimer(hWnd,PASTE_TIMER,PASTE_TIMEOUT,NULL);
            }
            break;
          case ID_COPY_COPYFIRSTCOLUMN:
            CopyFromListView(hwndListView,0);
            break;
          case ID_COPY_COPYENTIRELINE:
            CopyFromListView(hwndListView,1);
            break;
          case ID_FAV_ADD:
            ShowFavourites(hWnd);
            break;
          default:
            if (GET_WM_COMMAND_ID(wParam, lParam)>=UM_TOOLTIP)
            {
              ToggleOption(text,GET_WM_COMMAND_ID(wParam, lParam)-UM_TOOLTIP);
            }
            break;
        }
			  break;
      }
    case WM_TIMER:
      {
        DWORD exit_code;
        GetExitCodeThread(pthandle,&exit_code);
        if (exit_code!=STILL_ACTIVE)
        {
          TextWidget *text;

          text=(TextWidget*)GetWindowLong(hwndTextBox, 0);
          BuildToolbar(text,g_hInst,g_hWnd);
          KillTimer(hWnd,PASTE_TIMER);
        }
      }
      break;

		case WM_DESTROY:
			PostQuitMessage(0);
			break;

	  case WM_NOTIFY: 
      switch (((LPNMHDR) lParam)->code) 
      { 
        case TTN_GETDISPINFO: 
          { 
            LPNMTTDISPINFO lpttt; 
            TextWidget *text;

            text=(TextWidget*)GetWindowLong(hwndTextBox, 0);

            lpttt = (LPNMTTDISPINFO) lParam; 
            lpttt->hinst = NULL; 
            lpttt->uFlags=TTF_DI_SETITEM;

            switch (lpttt->hdr.idFrom) 
            { 
              case ID_FILE_CONNECT: 
                //lpttt->lpszText = "Logon"; 
                strcpy(lpttt->szText,"Logon");
                break; 
              case ID_NEW_WINDOW: 
                //lpttt->lpszText = "New Window"; 
                strcpy(lpttt->szText,"New Window");
                break; 
              case ID_COMMIT: 
                strcpy(lpttt->szText, "Commit"); 
                break; 
              case ID_ROLLBACK: 
                strcpy(lpttt->szText,"Rollback"); 
                break; 
              case ID_QUERY_CANCEL: 
                strcpy(lpttt->szText,"Cancel Query"); 
                break; 
              case ID_REPEAT: 
                strcpy(lpttt->szText,"Repeat Statement"); 
                break; 
              default:
                if (lpttt->hdr.idFrom>=UM_TOOLTIP)
                {
                  strcpy(lpttt->szText,default_options[lpttt->hdr.idFrom-UM_TOOLTIP].tooltip); 
                }
                break; 
            } 
            break; 
          } 
      }
      break;
		default:
			return DefWindowProc(hWnd, uMessage, wParam, lParam);
  }
  return 0;
}

BOOL InitApplication(HINSTANCE hInstance)
{
	WNDCLASSEX  wcex;
	ATOM        aReturn;

	wcex.cbSize          = sizeof(WNDCLASSEX);
	wcex.style           = 0;
	wcex.lpfnWndProc     = (WNDPROC)MainWndProc;
	wcex.cbClsExtra      = 0;
	wcex.cbWndExtra      = 0;
	wcex.hInstance       = hInstance;
	wcex.hCursor         = LoadCursor(NULL, IDC_ARROW);
	wcex.hbrBackground   = (HBRUSH)(COLOR_WINDOW + 1);
	wcex.lpszMenuName    = MAKEINTRESOURCE(IDR_MAIN_MENU);
	wcex.lpszClassName   = g_szClassName;
	wcex.hIcon           = LoadIcon(g_hInst, MAKEINTRESOURCE(IDI_JBSQL));
	wcex.hIconSm         = LoadImage(g_hInst, MAKEINTRESOURCE(IDI_JBSQL), IMAGE_ICON, 16, 16, 0);

	aReturn = RegisterClassEx(&wcex);

	if(0 == aReturn)
	{
		WNDCLASS wc;

		wc.style          = 0;
		wc.lpfnWndProc    = (WNDPROC)MainWndProc;
		wc.cbClsExtra     = 0;
		wc.cbWndExtra     = 0;
		wc.hInstance      = hInstance;
		wc.hIcon          = LoadIcon(g_hInst, MAKEINTRESOURCE(IDI_JBSQL));
		wc.hCursor        = LoadCursor(NULL, IDC_ARROW);
		wc.hbrBackground  = (HBRUSH)(COLOR_WINDOW + 1);
		wc.lpszMenuName   = MAKEINTRESOURCE(IDR_MAIN_MENU);
		wc.lpszClassName  = g_szClassName;

		aReturn = RegisterClass(&wc);
	}

	return aReturn;
}


BOOL InitInstance(HINSTANCE hInstance,
                  int nCmdShow)
{
	HWND     hWnd;
	TCHAR    szTitle[MAX_PATH] = "JB/SQL";

	g_hInst = hInstance;

	hWnd = CreateWindowEx(  0,
													g_szClassName,
													szTitle,
													WS_OVERLAPPEDWINDOW,
													CW_USEDEFAULT,
													CW_USEDEFAULT,
													CW_USEDEFAULT,
													CW_USEDEFAULT,
													NULL,
													NULL,
													hInstance,
													NULL);


	if (!hWnd)
		 return FALSE;

	ShowWindow(hWnd, nCmdShow);
	UpdateWindow(hWnd);
  g_hWnd=hWnd;
	return TRUE;

}

void CreateNewArgv(char ***argv,int *argc,char *commandline,int start_pos,int end_pos)
{
  *argv=(char**)realloc(*argv,(*argc+1)*sizeof(char*));
  (*argv)[*argc]=malloc(end_pos-start_pos+1);
  memcpy((*argv)[*argc],&commandline[start_pos],end_pos-start_pos);
  (*argv)[*argc][end_pos-start_pos]=0;
  (*argc)++;
}

void ConvertCommandLine(char *commandline,int *argc,char ***argv)
{
  int lastpos=0,x;
  char quote=0;

  *argv=NULL;
  *argc=0;
  for (x=0;x<=(int)strlen(commandline);x++)
  {
    if (!quote && (commandline[x]==' ' || commandline[x]=='\t' || commandline[x]==0))
    {
      if (lastpos!=x)
        CreateNewArgv(argv,argc,commandline,lastpos,x);
      lastpos=x+1;
    }
    else if ((commandline[x]=='"' || commandline[x]=='\'') && (!quote || commandline[x]==quote))
    {
      if (!quote)
      {
        quote=commandline[x];
      }
      else
      {
        CreateNewArgv(argv,argc,commandline,lastpos,x);
        quote=0;
      }
      if (quote==0 || x==0 || commandline[x-1]==' ' || commandline[x-1]=='\t')
        lastpos=x+1;
    }
  }
  *argv=(char**)realloc(*argv,(*argc+1)*sizeof(char*));
  (*argv)[*argc]=NULL;
}

void RunInitialCommand(TextWidget *text,int argc,char *argv[])
{
  int x;

  text->termout=0;
  ClearCommand(text,0,0);
  for (x=2;x<argc;x++)
  {
    if (x>2)
      InsertUnparsedLineBuffer(text," ");
    InsertUnparsedLineBuffer(text,argv[x]);
  }
  InsertNewLineBuffer(text);
  text->termout=1;
}

int PASCAL WinMain(HINSTANCE hInstance,
                   HINSTANCE hPrevInstance,
                   LPSTR lpCmdLine,
                   int nCmdShow)
{
	MSG  msg;
  char **argv;
  int argc;
  HACCEL acctable;
  HACCEL acctable2;
  HWND focus;

	g_hInst = hInstance;

	if(!hPrevInstance)
		if(!InitApplication(hInstance))
			return FALSE;

	InitCommonControls();

	if (!InitInstance(hInstance, nCmdShow))
		return FALSE;

  InitOCI();
  vht=CreateVariableHashTable();
  ConvertCommandLine(GetCommandLine(),&argc,&argv);
  strcpy(executable,argv[0]);
  tid=1;
  ReadCommandHistory(g_text);
  ReadUserCredentials();
  g_menu=LoadMenu(hInst,MAKEINTRESOURCE(IDR_LV_POPUP));
  if (argc>1)
  {
    tid=0;
    if (!ConnectWithConnectString(argv[1]))
    {
      tid=1;
      ShowPrompt(g_text);
    }
    else
      tid=1;
  }
  if (!ora_connected)
  {
	  if (!ShowLoginBox(g_hWnd,g_text))
		  return 0;
  }
  else
  {
    GetDefaultOptions(g_text);
    SetPWD(g_text);
  }
  EnableWindow(hwndTextBox,TRUE);
  tid=0;
  g_tid=GetCurrentThreadId();
  acctable=LoadAccelerators(hInstance,MAKEINTRESOURCE(IDR_ACCELERATOR));
  acctable2=LoadAccelerators(hInstance,MAKEINTRESOURCE(IDR_ACCEL_TW));

  if (argc>2)
    RunInitialCommand(g_text,argc,argv);

  while(GetMessage( &msg, NULL, 0x00, 0x00)>0)
	{
    if (msg.message==WM_KEYDOWN)
    {
      focus=GetFocus();
      if (g_text && focus==GetDlgItem(g_text->schema,ID_SCHEMABROWSER_CODE))
      {
        if (TranslateAccelerator(g_text->schema,acctable,&msg))
          continue;
      }
      else if (g_text && (focus==hwndListView || focus==hwndTextBox))
      {
        if (TranslateAccelerator(g_hWnd,acctable2,&msg))
          continue;
      }
    }
		TranslateMessage(&msg);
		DispatchMessage(&msg);
	}
  hwndTextBox=NULL;
  if (tid)
    AbortQuery();
  while (tid)
    Sleep(20);
  if (g_text->spoolfile)
    CloseHandle(g_text->spoolfile);
  SaveCommandHistory(g_text);
  CloseTMPFile();
	FreeCols(&cols,&ncols);
  if (ora_connected)
		DisconnectFromOracle();
	return msg.wParam;
}

