#include <windows.h>
#include <winbase.h>
#include <commctrl.h>
#include <stdio.h>
#include <stdarg.h>
#include "resource.h"
#include <oci.h>
#include <time.h>
#include "hashtab.h"

#define JBSQL_VERSION "1.0.2.8"

#define GET_WM_COMMAND_ID(wp, lp)               (wp)
#define GET_WM_COMMAND_HWND(wp, lp)             (HWND)(LOWORD(lp))
#define GET_WM_COMMAND_CMD(wp, lp)              HIWORD(lp)

#define ID_LISTVIEW       200
#define ID_EDIT           201
#define ID_SPLITTER       202
#define ID_HINTLIST       203
#define ID_TOOLBAR        204
#define ID_SCHEMABROWSER  205
#define ID_SQLSPLITTER    206
#define ID_SCHEMABROWSER_SCHEMA_TEXT 206
#define ID_SCHEMABROWSER_SCHEMA      207
#define ID_SCHEMABROWSER_TYPE_TEXT   208
#define ID_SCHEMABROWSER_TYPE        209
#define ID_SCHEMABROWSER_NAME_TEXT   210
#define ID_SCHEMABROWSER_NAME        211
#define ID_SCHEMABROWSER_CODE        212
#define ID_FAVOURITES_TEXT           213
#define ID_FAVOURITES_TREE           214
#define ID_FAVOURITES_PASTE          215
#define ID_FAVOURITES_CUT            216
#define ID_FAVOURITES_NEWFOLDER      217
#define ID_FAVOURITES_CLOSE          218


#define UM_TOOLTIP  1000 /* DO NOT CHANGE THIS!!! */

#define LEFT_TIMER    1
#define RIGHT_TIMER   2
#define TOP_TIMER     3
#define BOTTOM_TIMER  4
#define PASTE_TIMER   5

#define BUFLEN 5000
#define Y_OFFSET 2
#define X_OFFSET 2
#define CARET_OFFSET 2
#define SCROLL_TIMEOUT 20
#define PASTE_TIMEOUT 500
#define NTOKENS 10

#define DELIMITERS "()+-*/<>=!~;:.'@%,\\\"^&|{}?\t[] "
#define DELIMITERS_HINT "()+-*/<>=!~;:'@%,\\\"^&|{}?\t[]\n "
#define MOVE_DELIMITERS "\n()+-*/<>=!~;:.'@%,\\\"^&|{}?\t[] "

#define MIN(x,y) (((int)(x)>(int)(y))?(y):(x))
#define MAX(x,y) (((int)(x)<(int)(y))?(y):(x))

#define STR_OPTION(o) text->options[o].value.str_value
#define INT_OPTION(o) text->options[o].value.int_value

typedef int (*QueryProc)(HWND hWnd,char *schema,char *object);
typedef int (*SetOptionProc)(void *text,void *value);


typedef struct {
  char  *col_name;
  int    col_size;
  union {
    char          *str_value;
    OCILobLocator **lob_value;
  } value;
  short *indicator;
  int   datatype;
} ColumnDesc;

typedef enum { ALLOW_NEWLINE=0,
               ARRAYSIZE,
               AUTOCOMPLETE,
               AUTOTRACE,
               COMMANDHISTORY,
               DEBUG,
               DIRECTORY,
               EDITOR,
               EDIT_FILENAME,
               FEEDBACK,
               HEADING,
               HISTORY,
               LISTVIEW,
               LONG_SIZE,
               MAX_SCRIPTS,
               MAXLINES,
               MAXMEM,
               PROMPT,
               SAVEHISTORY,
               SCAN,
               SCREEN,
               SERVEROUTPUT,
               SOUND,
               SOUNDFILE,
               SOUNDTIME,
               TIMING,
               VERIFY
};

typedef enum { O_CHAR, O_INT, O_BOOLEAN, O_DIR } OptionType;
typedef enum { V_CHAR, V_NUMBER, V_DATE } VariableType;

typedef struct {
  char option[30];
  union {
    char *str_value;
    DWORD int_value;
  } value;
  int value_size;
  int icon;
  int on_toolbar;
  OptionType type;
  char *tooltip;
  char *description;
  SetOptionProc proc;
} Option;

typedef struct {
  int  statistic;
  int  value;
  char name[65];
} SessionStatistic;

typedef struct {
  char *command;
  char lastchar;
  int  sqlcommand;
} CommandHistory;

typedef struct {
  char  **text;    /* Contains all text lines */
	int   lines;     /* Contains the number of lines in the text widget */
	int   top;       /* Contains the first displayed line number */
	int   left;      /* Contains the first displayed left character */
	int   width;     /* Contains the number of visible characters */
	int   height;    /* Contains the number of visible lines */
  int   f_width;   /* The font width */
  int   f_height;  /* The font height */
  int   w_width;   /* The window width */
  int   w_height;  /* The window height */
  int   cur_pos;   /* Current cursor position */
  int   cur_line;  /* Current line number, 1-based */
  HDC   hdc;       /* The device context */
  HFONT hfont;     /* The font to draw with */
  int   active;    /* Is the window active? */
  HWND  hintlist;  /* The hintlist window */
  int   selection; /* Is there an active selection */
  int   sel_left;  /* Left selection position */
  int   sel_top;   /* Top selection position */
  int   sel_right; /* Right selection position */
  int   sel_bottom;/* Bottom selection position */
  int   no_process;/* Do not process newlines */
  char  pwd[MAX_PATH]; /* The current directory */
  char  dir[MAX_PATH]; /* The script directory */
  int   termout;   /* Output commands to terminal */
  int   noredraw;  /* Must we redraw the screen on WPrintf */
  int   scripts;   /* The number of scripts currently nested */
  Option *options; /* The current options */
  DWORD start_time;/* The start time of a query */
  HANDLE spoolfile;/* The file to spool to */
  UINT  leftselect;  /* Timer for left offscreen mouse selection */
  UINT  rightselect; /* Timer for right offscreen mouse selection */
  UINT  topselect;   /* Timer for top offscreen mouse selection */
  UINT  bottomselect;/* Timer for bottom offscreen mouse selection */
  int   scanf;     /* Are we asking for user input? 0=no, 1=yes, 2=yes, but hidden */
  int   old_screen;/* temporary storage for the screen option when running scripts */
  HWND  analyze;   /* The handle of the sql analyzer window */
  HWND  schema;    /* The handle of the schema browser window */
  SessionStatistic *pre_stats; /* Pre statement statistics */
  int    npre_stats; /* The number of pre-statistics */
  HWND  toolbar;   /* The window handle of the toolbar */
  CommandHistory work_cmd; /* The command-line that the user was working on when he presses the up arrow */
  int   cancelled; /* Has the execution of a script been cancelled */
  char *(tokens[10]);
  int   ntokens;   /* The number of tokens in the command */
} TextWidget;

typedef struct {
  TextWidget *text;
  char       *data;
  char       action[500];
  char       executestr[500];
} ExecuteStruct;

typedef struct {
  char *token;
  char *prev_token;
  char *query;
  char *values;
} SQLToken;

typedef struct {
  char *keyword;
  char *value;
} Keyword;  

typedef struct {
  int       id;
  HTREEITEM tree_item;
} PlanElement;

typedef struct {
  char hint[40];
  int  table;
  int  index;
  char *description;
} SQLHint;



extern HINSTANCE hInst,g_hInst;
extern HWND g_hWnd;
extern HWND hwndListView;
extern HWND hwndTextBox;
extern HWND hwndSplitter;
extern HWND g_toolbar;
extern HMENU g_menu;
extern int ora_connected;
extern int recordlength;
extern HANDLE f;
extern ColumnDesc *cols;
extern int ncols;
extern int nrows;
extern int total_qrows;
extern int linepos;
extern int nlines;
extern DWORD tid;
extern DWORD ptid;
extern char *command;
extern int allow_blank_lines;
extern TextWidget *g_text;
extern CommandHistory *old_commands;
extern int nold_commands;
extern char last_usernames[10][100];
extern char last_databases[10][100];
extern int noptions;
extern Option default_options[];
extern char connectstring[MAX_PATH],g_database[250];
extern HashTable vht;
extern DWORD g_tid;
extern int max_command_size,cur_command_size;
extern int window_h_offset,window_v_offset;

void ReallocCommand(int size);
void *MallocCommand(int size);
void FreeCommand(void);
char *StrdupCommand(char *str);

void Error(HWND hWnd,char *fmt,...);
void DisplayLastError(HWND hWnd);
void Log(char *fmt,...);
char *myvsprintf(const char *fmt, va_list vl);
void WPrintf(char *fmt,...);
void WPrintfUP(char *str);
void WPrintfNL(void);
void WPrintfOnly(char *fmt,...);
void FPrintf(char *fmt,...);
void WScanf(TextWidget *text,char *str,int hidden);
int ReplaceString(char *str,char *srch,char *repl);
void ConvertCommandLine(char *commandline,int *argc,char ***argv);

HWND CreateListView(HINSTANCE hInstance, HWND hwndParent);
void ResizeListView(HWND hwndListView, HWND hwndParent);
BOOL InitListView(HWND hwndListView);
LRESULT ListViewNotify(HWND hWnd, LPARAM lParam);
void ClearList(HWND hwndListView);

HWND CreateTextWidget(HINSTANCE hInst,HWND parent,TextWidget **rtext);
void InsertLineBuffer(TextWidget *text,char *fmt,...);
void InsertUnparsedLineBuffer(TextWidget *text,char *str);
void InsertNewLineBuffer(TextWidget *text);
void UpdateCaretPosition(TextWidget *text,int scroll);
void RedrawLine(TextWidget *text,int line);
void RefreshTextWidget(TextWidget *text);

HWND CreateSplitter(HINSTANCE hInstance, HWND hwndParent);
void ResizeSplitter(HWND hwndSplitter, HWND hwndParent);

void EnableThreads(void);
int InitOCI(void);

int ConnectToOracle(char *username,char *password,char *database,char *connect_as);
int DisconnectFromOracle(void);
int ConnectWithConnectString(char *connectstring);
void AbortQuery(void);
void SetWindowTitle(char *username,char *database);


void ShowPrompt(TextWidget *text);
void InsertCommandLine(char *str);
int ProcessCommand(TextWidget *text);
void PrintErrorCharacter(int c);
void PreviousStatement(TextWidget *text);
void NextStatement(TextWidget *text);
void RedrawCommand(TextWidget *text,int pos,int nlines);

int CreateTMPFile(TextWidget *text);
void CloseTMPFile(void);

int PrepareSource(TextWidget *text,char *p_sqlstr,ColumnDesc **p_cols,int *p_ncols,char *action);
int FetchData(TextWidget *text,ColumnDesc *cols,int ncols,char *action);
int FinishSource(ColumnDesc *cols,int ncols);
void FreeCols(ColumnDesc **cols,int *ncols);

int DoCommand(char *str,int feedback,char *executestr);
int ExecuteQuery(ExecuteStruct *s);
void ClearCommand(TextWidget *text,int add,int sql_command);

void Commit(void);
void Rollback(void);
void GetUserID(char *object,char **user_id,char **object_name,int show_error);
char *GetOracleObjectType(char *owner,char *object,int show_error);
int ExecuteDDLStatement(char *str);

int ShowHint(TextWidget *text,char *command,int linepos);
void DestroyHintList(TextWidget *text);
void SelectFromHint(TextWidget *text,int linepos);
void FillHintListFromQuery(TextWidget *text,char *query,char *name,char *owner,char *object_type);
void AddToHintList(TextWidget *text,char *str);

void Describe(TextWidget *text,char *command);
void ShowError(TextWidget *text,char *type,char *name);

char *Trim(char *str);

void RemoveLines(TextWidget *text,int nlines);
void SyncTextPosition(TextWidget *text);
void AppendLineBuffer(char *str);
int DeleteLineBufferCharacter(TextWidget *text);
int DeleteNextLineBufferCharacter(TextWidget *text);
void InsertTab(TextWidget *text);
void InsertNewLine(TextWidget *text);
void MoveCursorLeft(TextWidget *text,int next_word);
void MoveCursorRight(TextWidget *text,int next_word);
void MoveCursorHome(TextWidget *text,int full);
void MoveCursorEnd(TextWidget *text,int full);
void MoveCursorUp(TextWidget *text);
void MoveCursorDown(TextWidget *text);
void AppendLine(TextWidget *text,char *str);
void InsertLine(TextWidget *text);
void ClearLine(TextWidget *text);
void DeleteCharacter(TextWidget *text);
void DeleteNextCharacter(TextWidget *text);
void GetRealSelection(TextWidget *text,int *sel_left,int *sel_top,int *sel_right,int *sel_bottom);

int RunScript(ExecuteStruct *s);

BOOL FAR PASCAL OptionProc(HWND hWnd, unsigned message,DWORD wParam, LONG lParam);
int GetDefaultOptions(TextWidget *text);
void SetCommandLineOption(TextWidget *text,char *settings);
void SetPWD(TextWidget *text);
void PrintOptions(TextWidget *text);
void SaveUserCredentials(char *username,char *database);
void ReadUserCredentials(void);
void ToggleOption(TextWidget *text,int option);
void CopyOptions(Option *source,Option *target);
void FreeOptions(Option *source);
int EnableServerOutput(void *text,void *value);

void WriteRegistryString(char *key,char *subkey,char *value);
void WriteRegistryInteger(char *key,char *subkey,DWORD value);
void WriteRegistryBytes(char *key,char *subkey,void *value,int sz);
BOOLEAN ReadRegistryString(char *key,char *subkey,char *value,int sz);
BOOLEAN ReadRegistryInteger(char *key,char *subkey,DWORD *value);

int PrintServerOutput(TextWidget *text);
void SaveCommandHistory(TextWidget *text);
void ReadCommandHistory(TextWidget *text);

HWND BuildToolbar(TextWidget *text,HINSTANCE hInst,HWND hWnd);

HashTable CreateVariableHashTable(void);
void DefineVariable(char *str);
void AcceptVariable(TextWidget *text,char **tokens,int ntokens);
int CheckVariableFormat(VariableType vt,char *format,char *value);

HWND ShowSqlAnalyzer(HWND parenthWnd);
void ExplainPlan(HWND hWnd,char *sql_statement);
void FillListBoxFromQuery(HWND hWnd,char *query,char **bindvar,int nbindvars,int listbox);
void ShowAutotrace(ExecuteStruct *s);
void AnalyzeError(char *fmt,...);
void SetAnalyzerText(TextWidget *text,char *sqlstmt);

void ShowStatistics(TextWidget *text);
void GetStatistics(SessionStatistic **s,int *ns);

void DeleteCharacterFromString(char *str,int x);

char *GetCurrentUser(void);
HWND ShowSchemaBrowser(HWND parent);
void SchemaPrintf(HWND hWnd,char *fmt,...);
int GenerateSource(HWND hWnd,char *schema,char *name,char *type);
int ProcessQuery(HWND hWnd,char *query,char *schema,char *name,QueryProc do_function);
int ProcessTable(HWND hWnd,char *schema,char *table);
int ProcessIndex(HWND hWnd,char *schema,char *index);
int ProcessConstraint(HWND hWnd,char *schema,char *constraint);
int ProcessFunction(HWND hWnd,char *schema,char *function);
int ProcessProcedure(HWND hWnd,char *schema,char *procedure);
int ProcessPackage(HWND hWnd,char *schema,char *package);
int ProcessPackageBody(HWND hWnd,char *schema,char *packagebody);
int ProcessTrigger(HWND hWnd,char *schema,char *trigger);
int ProcessView(HWND hWnd,char *schema,char *view);
int ProcessType(HWND hWnd,char *schema,char *type);
int ProcessTypeBody(HWND hWnd,char *schema,char *type);
int ProcessJavaSource(HWND hWnd,char *schema,char *javasource);
int ProcessSynonym(HWND hWnd,char *schema,char *synonym);
int ProcessSequence(HWND hWnd,char *schema,char *sequence);

void ExportData(TextWidget *text);
int GetSingleValue(char *query,char *target,int target_size);
void CopyFromListView(HWND hWnd,int entire_line);
void CopyToClipboard(TextWidget *text);
int PasteFromClipboard(TextWidget *text);

void SaveToHistory(char *command);

HWND ShowFavourites(HWND parent);

