#include "jboci.h"

int ExecuteDDL(char *str)
{
  ub4 qrows;

  if (!ora_connected)
  {
    Error(NULL,"Not connected to a database\n");
    return -1;
  }
  if (ChkOCIErr2(OCIHandleAlloc(g_oci->envhp, 
                               &g_oci->stmtp,
                               OCI_HTYPE_STMT, 
                               0, 
                               NULL)))
  {
    return -1;
  }
  if (ChkOCIErr2(OCIStmtPrepare(g_oci->stmtp,
                               g_oci->errhp,
                               str,
                               strlen(str),
                               OCI_NTV_SYNTAX,
                               OCI_DEFAULT)))
  {
    OCIHandleFree(g_oci->stmtp,OCI_HTYPE_STMT);
    g_oci->stmtp=NULL;
    return -2;
  }
  g_oci->offset=1;
  if (ChkOCIErr2(OCIStmtExecute(g_oci->svchp,
                               g_oci->stmtp,
                               g_oci->errhp,
                               1,
                               0,
                               NULL,
                               NULL,
                               OCI_DEFAULT)))
  {
    g_oci->offset=0;
    OCIHandleFree(g_oci->stmtp,OCI_HTYPE_STMT);
    g_oci->stmtp=NULL;
    return -3;
  }
  g_oci->offset=0;
  ChkOCIErr2(OCIAttrGet(g_oci->stmtp,
                           OCI_HTYPE_STMT, 
                           &qrows,
                           NULL,
                           OCI_ATTR_ROW_COUNT, 
                           g_oci->errhp));
  OCIHandleFree(g_oci->stmtp,OCI_HTYPE_STMT);
  g_oci->stmtp=NULL;
  return qrows; 
}


void ExplainPlan(HWND hWnd,char *sql_statement)
{
  char *str;
  char char_ptr[2000],stmt_id[50];
  int len=2000,result;
  int id,parent_id;
  char query[700]="";
  OCIBind *bndhp=NULL;
  OCIDefine *defhp=NULL;
  HashTable ht;
  TVINSERTSTRUCT item_data;
  HTREEITEM tree_item;
  PlanElement pe;
  ub2 ind1;
  time_t t;

  sprintf(stmt_id,"JBSQL_%l",(long)time(&t));
  TreeView_DeleteAllItems(GetDlgItem(GetDlgItem(hWnd,ID_SQLSPLITTER),IDC_PLAN_TREE));
  ExecuteDDL("savepoint JBSQL");
  sprintf(char_ptr,"delete from plan_table where statement_id='%s'",stmt_id);
  if (ExecuteDDL(char_ptr)<0)
  {
    Error(NULL,"Can't access the plan table, maybe you need to run the utlxplan.sql script");
    return;
  }
  str=(char*)malloc(strlen(sql_statement)+250);
  if (sql_statement[strlen(sql_statement)-1]==';')
    sql_statement[strlen(sql_statement)-1]='\0';
  sprintf(str,"EXPLAIN PLAN SET STATEMENT_ID = '%s' FOR %s",stmt_id,sql_statement);
  if (ExecuteDDL(str)<0)
  {
    free(str);
    return;
  }
  free(str);

  sprintf(query,"SELECT ID,PARENT_ID,OPERATION||DECODE(OTHER_TAG,NULL,'','*')||DECODE(OPTIONS,NULL,'',' ('||OPTIONS||')')||DECODE(OBJECT_NAME,NULL,'',' OF '''||OBJECT_NAME||'''')||DECODE(OBJECT_TYPE,NULL,'',' ('||OBJECT_TYPE||')')||DECODE(ID,0,DECODE(OPTIMIZER,NULL,'',' Optimizer='||OPTIMIZER))||DECODE(COST,NULL,'',' (Cost='||COST||DECODE(CARDINALITY,NULL,'',' Card='||CARDINALITY)||DECODE(BYTES,NULL,'',' Bytes='||BYTES||')'))||DECODE(PARTITION_START,NULL,'',' PStart='||PARTITION_START)||DECODE(PARTITION_STOP,NULL,'',' PStop='||PARTITION_STOP) PLAN FROM PLAN_TABLE START WITH ID=0 AND STATEMENT_ID='%s' CONNECT BY PRIOR ID=PARENT_ID AND STATEMENT_ID='%s' ORDER BY ID,POSITION",stmt_id,stmt_id);
  char_ptr[0]='\0';
  if (ChkOCIErr2(OCIHandleAlloc(g_oci->envhp, 
                               &g_oci->stmtp,
                               OCI_HTYPE_STMT, 
                               0, 
                               NULL)))
  {
    return;
  }
  if (ChkOCIErr2(OCIStmtPrepare(g_oci->stmtp,
                               g_oci->errhp,
                               query,
                               strlen(query),
                               OCI_NTV_SYNTAX,
                               OCI_DEFAULT)))
  {
    OCIHandleFree(g_oci->stmtp,OCI_HTYPE_STMT);
    g_oci->stmtp=NULL;
    return;
  }
  if (ChkOCIErr2(OCIDefineByPos(g_oci->stmtp,
                               &defhp,
                               g_oci->errhp,
                               1,
                               &id,
                               sizeof(int),
                               SQLT_INT,
                               NULL,
                               NULL,
                               NULL,
                               OCI_DEFAULT)))
  {
    OCIHandleFree(g_oci->stmtp,OCI_HTYPE_STMT);
    g_oci->stmtp=NULL;
    return;
  }

  if (ChkOCIErr2(OCIDefineByPos(g_oci->stmtp,
                               &defhp,
                               g_oci->errhp,
                               2,
                               &parent_id,
                               sizeof(int),
                               SQLT_INT,
                               &ind1,
                               NULL,
                               NULL,
                               OCI_DEFAULT)))
  {
    OCIHandleFree(g_oci->stmtp,OCI_HTYPE_STMT);
    g_oci->stmtp=NULL;
    return;
  }

  if (ChkOCIErr2(OCIDefineByPos(g_oci->stmtp,
                               &defhp,
                               g_oci->errhp,
                               3,
                               char_ptr,
                               len,
                               SQLT_STR,
                               NULL,
                               NULL,
                               NULL,
                               OCI_DEFAULT)))
  {
    OCIHandleFree(g_oci->stmtp,OCI_HTYPE_STMT);
    g_oci->stmtp=NULL;
    return;
  }

  g_oci->offset=0;
  result=OCIStmtExecute(g_oci->svchp,
                        g_oci->stmtp,
                        g_oci->errhp,
                        0,
                        0,
                        NULL,
                        NULL,
                        OCI_DEFAULT);
  if (result!=OCI_SUCCESS && result!=OCI_NO_DATA)
  {
    g_oci->offset=0;
    OCIHandleFree(g_oci->stmtp,OCI_HTYPE_STMT);
    g_oci->stmtp=NULL;
    return;
  }
  ht=HTCreate(101,sizeof(int),sizeof(PlanElement),NULL,NULL,NULL,NULL);
  for (;;)
  {
    result=OCIStmtFetch(g_oci->stmtp,
                        g_oci->errhp,
                        1,
                        OCI_FETCH_NEXT,
                        OCI_DEFAULT);
    if (result!=OCI_SUCCESS && result!=OCI_NO_DATA && result!=OCI_SUCCESS_WITH_INFO)
    {
      break;
    }
    if (result==OCI_NO_DATA)
      break;

    if (ind1==(ub2)-1)
      item_data.hParent=TVI_ROOT;
    else
    {
      HTGetRecord(ht,&parent_id,&pe);
      item_data.hParent=pe.tree_item;
    }
    item_data.hInsertAfter=TVI_LAST;
    item_data.item.mask=TVIF_STATE|TVIF_TEXT;
    item_data.item.state=TVIS_EXPANDED;
    item_data.item.stateMask=TVIS_EXPANDED;
    item_data.item.pszText=char_ptr;
    item_data.item.cchTextMax=strlen(char_ptr);

    tree_item=TreeView_InsertItem(GetDlgItem(GetDlgItem(hWnd,ID_SQLSPLITTER),IDC_PLAN_TREE),&item_data);
    pe.id=id;
    pe.tree_item=tree_item;
    HTAddRecord(ht,&pe);
  }
  HTDestroy(ht);
  g_oci->offset=0;
  OCIHandleFree(g_oci->stmtp,OCI_HTYPE_STMT);
  g_oci->stmtp=NULL;
  ExecuteDDL("rollback to savepoint JBSQL");
}

void FillListBoxFromQuery(HWND hWnd,char *query,char **bindvar,int nbindvars,int listbox)
{
  char *char_ptr=NULL;
  int len=40,result,x;
  OCIBind *bndhp=NULL;
  OCIDefine *defhp=NULL;
  ub4 qrows;
  int nrows;
  char query_finished=0;
  ub4 total_qrows=0;
  int fetch_array_size=SCHEMA_ARRAY_SIZE;

  if (tid)
    return;
  if (ChkOCIErr2(OCIHandleAlloc(g_oci->envhp, 
                               &g_oci->stmtp,
                               OCI_HTYPE_STMT, 
                               0, 
                               NULL)))
  {
    return;
  }
  if (ChkOCIErr2(OCIStmtPrepare(g_oci->stmtp,
                               g_oci->errhp,
                               query,
                               strlen(query),
                               OCI_NTV_SYNTAX,
                               OCI_DEFAULT)))
  {
    OCIHandleFree(g_oci->stmtp,OCI_HTYPE_STMT);
    g_oci->stmtp=NULL;
    return;
  }
  for (x=0;x<nbindvars;x++)
  {
    if (ChkOCIErr2(OCIBindByPos(g_oci->stmtp,
                               &bndhp,
                               g_oci->errhp,
                               x+1,
                               bindvar[x],
                               strlen(bindvar[x]),
                               SQLT_CHR,
                               NULL,
                               NULL,
                               NULL,
                               0,
                               NULL,
                               OCI_DEFAULT)))
    {
      OCIHandleFree(g_oci->stmtp,OCI_HTYPE_STMT);
      g_oci->stmtp=NULL;
      return;
    }
  }
  char_ptr=(char*)malloc(SCHEMA_ARRAY_SIZE*len*sizeof(char));
  if (char_ptr==NULL)
  {
    Error(NULL,"Out of memory");
    return;
  }
  if (ChkOCIErr2(OCIDefineByPos(g_oci->stmtp,
                               &defhp,
                               g_oci->errhp,
                               1,
                               char_ptr,
                               len,
                               SQLT_STR,
                               NULL,
                               NULL,
                               NULL,
                               OCI_DEFAULT)))
  {
    OCIHandleFree(g_oci->stmtp,OCI_HTYPE_STMT);
    g_oci->stmtp=NULL;
    free(char_ptr);
    return;
  }

  g_oci->offset=0;
  result=OCIStmtExecute(g_oci->svchp,
                        g_oci->stmtp,
                        g_oci->errhp,
                        0,
                        0,
                        NULL,
                        NULL,
                        OCI_DEFAULT);
  if (result!=OCI_SUCCESS && result!=OCI_NO_DATA)
  {
    g_oci->offset=0;
    OCIHandleFree(g_oci->stmtp,OCI_HTYPE_STMT);
    g_oci->stmtp=NULL;
    free(char_ptr);
    return;
  }
  while (!query_finished)
  {
    result=OCIStmtFetch(g_oci->stmtp,
                        g_oci->errhp,
                        fetch_array_size,
                        OCI_FETCH_NEXT,
                        OCI_DEFAULT);
    if (result!=OCI_SUCCESS && result!=OCI_NO_DATA && result!=OCI_SUCCESS_WITH_INFO)
    {
      if (ChkOCIErrI(result,1406))
        break;
    }
    if (ChkOCIErr2(OCIAttrGet(g_oci->stmtp,
                             OCI_HTYPE_STMT, 
                             &qrows,
                             NULL,
                             OCI_ATTR_ROW_COUNT, 
                             g_oci->errhp)))
    {
      break;
    } 
    nrows = qrows - total_qrows;
    if (nrows!=fetch_array_size)
      query_finished=1;

    for (x=0;x<(int)(qrows-total_qrows);x++)
    {
      SendMessage(hWnd,listbox?LB_ADDSTRING:CB_ADDSTRING,0,(LPARAM)&char_ptr[x*len]);
    }
    total_qrows=qrows;
  }
  g_oci->offset=0;
  OCIHandleFree(g_oci->stmtp,OCI_HTYPE_STMT);
  g_oci->stmtp=NULL;
  free(char_ptr);
}