#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "jboci.h"

int fetch_array_size;
HANDLE f=NULL;
char *lob_buffer=NULL;
int query_finished=1,total_qrows=0,long_column=0;

int CreateTMPFile(TextWidget *text)
{
  char tmppath[MAX_PATH],tmpfilename[MAX_PATH];

  CloseTMPFile();
  if (!INT_OPTION(LISTVIEW))
    return 0;
  if (GetTempPath(MAX_PATH,tmppath)==0)
  {
    DisplayLastError(NULL);
    return -1;
  }
  if (GetTempFileName(tmppath,"jbsql",0,tmpfilename)==0)
  {
    DisplayLastError(NULL);
    return -1;
  }
  f=CreateFile(tmpfilename,
             GENERIC_READ|GENERIC_WRITE,
             FILE_SHARE_READ,
             NULL,
             TRUNCATE_EXISTING,
             FILE_ATTRIBUTE_TEMPORARY|FILE_FLAG_DELETE_ON_CLOSE,
             NULL);
             
  if (f==INVALID_HANDLE_VALUE)
  {
    DisplayLastError(NULL);
    WPrintf("Failed to create tmp file\n");
    f=NULL;
    return -1;
  }
  return 0;
}

void CloseTMPFile(void)
{
  if (f!=NULL)
    CloseHandle(f);
  f=NULL;
}

void PrintHeaders(TextWidget *text,ColumnDesc *cols,int ncols)
{
  int x;
  char *lines;

  if (INT_OPTION(HEADING))
  {
    if (INT_OPTION(SCREEN))
    {
      text->noredraw=1;
      WPrintf("\r");
      for (x=0;x<(int)ncols;x++)
      {
        if (x)
          WPrintf(" %-*s",MAX(cols[x].col_size-1,strlen(cols[x].col_name)),cols[x].col_name);
        else
          WPrintf("%-*s",MAX(cols[x].col_size-1,strlen(cols[x].col_name)),cols[x].col_name);
      }
      WPrintfNL();
      for (x=0;x<(int)ncols;x++)
      {
        if (x)
          WPrintf(" ");
        lines=(char*)malloc(MAX(cols[x].col_size-1,(int)strlen(cols[x].col_name))+1);
        memset(lines,'-',MAX(cols[x].col_size-1,(int)strlen(cols[x].col_name)));
        lines[MAX(cols[x].col_size-1,(int)strlen(cols[x].col_name))]='\0';
        WPrintf("%s",lines);
        free(lines);
      }
      text->noredraw=0;
      WPrintfNL();
    }
    else if (text->spoolfile!=NULL)
    {
      for (x=0;x<(int)ncols;x++)
      {
        if (x)
          FPrintf(" %-*s",MAX(cols[x].col_size-1,strlen(cols[x].col_name)),cols[x].col_name);
        else
          FPrintf("%-*s",MAX(cols[x].col_size-1,strlen(cols[x].col_name)),cols[x].col_name);
      }
      FPrintf("\n");
      for (x=0;x<(int)ncols;x++)
      {
        if (x)
          FPrintf(" ");
        lines=(char*)malloc(MAX(cols[x].col_size-1,(int)strlen(cols[x].col_name))+1);
        memset(lines,'-',MAX(cols[x].col_size-1,(int)strlen(cols[x].col_name)));
        lines[MAX(cols[x].col_size-1,(int)strlen(cols[x].col_name))]='\0';
        FPrintf("%s",lines);
        free(lines);
      }
      FPrintf("\n");
    }
  }
}

int PrepareSource(TextWidget *text,char *p_sqlstr,ColumnDesc **p_cols,int *p_ncols,char *action)
{
  int        x,y;
  OCIParam   *param;
  ub2        dtype;
  char       *col_name;
  ub4        col_name_len;
  ub4        ncols;
  ColumnDesc *pcols;
  ub2        len;
  OCIDefine  *defhp=NULL;
  char       *char_ptr;
  int        record_size=0;
  int        total_memory;

  if (!ora_connected)
  {
//    WPrintf("\nNot connected to a database\n");
    WPrintf("Not connected to a database\n");
    return -1;
  }
  if (ChkOCIErr(OCIHandleAlloc(g_oci->envhp, 
                               &g_oci->stmtp,
                               OCI_HTYPE_STMT, 
                               0, 
                               NULL)))
  {
    return -1;
  }
  if (action[0]!='\0' && INT_OPTION(FEEDBACK))
  {
    HANDLE tmp=text->spoolfile;

    text->spoolfile=NULL;
  	WPrintfOnly("\rParsing statement...");
    text->spoolfile=tmp;
  }
  if (ChkOCIErr(OCIStmtPrepare(g_oci->stmtp,
                               g_oci->errhp,
                               p_sqlstr,
                               strlen(p_sqlstr),
                               OCI_NTV_SYNTAX,
                               OCI_DEFAULT)))
  {
    OCIHandleFree(g_oci->stmtp,OCI_HTYPE_STMT);
    g_oci->stmtp=NULL;
    return -2;
  }
  g_oci->offset=1;
  if (action[0]!='\0' && INT_OPTION(FEEDBACK))
  {
    HANDLE tmp=text->spoolfile;

    text->spoolfile=NULL;
  	WPrintfOnly("\rExecuting query...");
    text->spoolfile=tmp;
  }
  if (ChkOCIErr(OCIStmtExecute(g_oci->svchp,
                               g_oci->stmtp,
                               g_oci->errhp,
                               0,
                               0,
                               NULL,
                               NULL,
                               OCI_DEFAULT)))
  {
    g_oci->offset=0;
    OCIHandleFree(g_oci->stmtp,OCI_HTYPE_STMT);
    g_oci->stmtp=NULL;
    return -3;
  }

  

  g_oci->offset=0;
  if (ChkOCIErr(OCIAttrGet(g_oci->stmtp, 
                           OCI_HTYPE_STMT, 
                           &ncols, 
                           NULL, 
                           OCI_ATTR_PARAM_COUNT, 
                           g_oci->errhp)))
  {
    OCIHandleFree(g_oci->stmtp,OCI_HTYPE_STMT);
    g_oci->stmtp=NULL;
    return -4;
  }

  if ((pcols=(ColumnDesc*)malloc(sizeof(ColumnDesc)*ncols))==NULL)
  {
    WPrintf("Can't perform malloc!\n");
    return 1;
  }
  for (x=0;x<(int)ncols;x++)
  {
    if (ChkOCIErr(OCIParamGet(g_oci->stmtp, 
                              OCI_HTYPE_STMT, 
                              g_oci->errhp, 
                              &param,
                              x+1)))
    {
      break;
    }
    if (ChkOCIErr(OCIAttrGet(param,
                             OCI_DTYPE_PARAM, 
                             &dtype,
                             0,
                             OCI_ATTR_DATA_TYPE, 
                             g_oci->errhp)))
    {
      break;
    }

    if (ChkOCIErr(OCIAttrGet(param,
                             OCI_DTYPE_PARAM, 
                             &col_name,
                             &col_name_len,
                             OCI_ATTR_NAME, 
                             g_oci->errhp)))
    {
      break;
    }
    pcols[x].col_name=malloc(col_name_len+1);
    memcpy(pcols[x].col_name,col_name,col_name_len);
    pcols[x].col_name[col_name_len]='\0';
    if (ChkOCIErr(OCIAttrGet(param,
                             OCI_DTYPE_PARAM, 
                             &len,
                             NULL,
                             OCI_ATTR_DATA_SIZE, 
                             g_oci->errhp)))
    {
      break;
    } 
    pcols[x].datatype=dtype;
    switch (dtype)
    {
      case OCI_TYPECODE_REF:
        WPrintf("Unsupported datatype: REF for column %s\n",pcols[x].col_name);
        return -5;
      case OCI_TYPECODE_DATE:
      case OCI_TYPECODE_TIMESTAMP:
      case OCI_TYPECODE_TIME:
      case OCI_TYPECODE_TIME_TZ:
      case OCI_TYPECODE_TIMESTAMP_TZ:
        pcols[x].col_size=40;
        break;
      case OCI_TYPECODE_REAL:
      case OCI_TYPECODE_DOUBLE:
      case OCI_TYPECODE_FLOAT:
      case OCI_TYPECODE_NUMBER:
      case OCI_TYPECODE_DECIMAL:
      case OCI_TYPECODE_OCTET:
      case OCI_TYPECODE_INTEGER:
      case OCI_TYPECODE_SMALLINT:
        pcols[x].col_size=40;
        break;
      case OCI_TYPECODE_RAW:
      case OCI_TYPECODE_VARCHAR2:
      case OCI_TYPECODE_VARCHAR:
      case OCI_TYPECODE_CHAR:
        pcols[x].col_size=len+1;
        break;
      case OCI_TYPECODE_VARRAY:
        WPrintf("Unsupported datatype: VARRAY for column %s\n",pcols[x].col_name);
        return -5;
      case OCI_TYPECODE_TABLE:
        WPrintf("Unsupported datatype: TABLE for column %s\n",pcols[x].col_name);
        return -5;
      case OCI_TYPECODE_CLOB:
        pcols[x].col_size=INT_OPTION(LONG_SIZE)+1;
        break;
      case OCI_TYPECODE_BLOB:
        pcols[x].col_size=INT_OPTION(LONG_SIZE)+1;
        break;
      case OCI_TYPECODE_BFILE:
        pcols[x].col_size=INT_OPTION(LONG_SIZE)+1;
        break;
      case OCI_TYPECODE_OBJECT:
        WPrintf("\rUnsupported datatype: OBJECT for column %s\n",pcols[x].col_name);
        return -5;
      case OCI_TYPECODE_NAMEDCOLLECTION:
        WPrintf("\rUnsupported datatype: COLLECTION for column %s\n",pcols[x].col_name);
        return -5;
      case SQLT_LNG:
        pcols[x].col_size=INT_OPTION(LONG_SIZE)+1;
        break;
      case SQLT_BIN:
      case SQLT_LBI:
        pcols[x].col_size=INT_OPTION(LONG_SIZE)+1;
        break;
      case SQLT_RID:
      case SQLT_RDD:
        pcols[x].col_size=40;
        break;
      default:
        WPrintf("\rUnknown datatype %d for column %s\n",dtype,pcols[x].col_name);
        return -5;
    }
    record_size+=pcols[x].col_size;
  }
  total_memory=INT_OPTION(ARRAYSIZE)*record_size;
  if (total_memory<0 || total_memory>(int)INT_OPTION(MAXMEM))
  {
    fetch_array_size=INT_OPTION(MAXMEM)/record_size;
    if (fetch_array_size<1)
      fetch_array_size=1;
  }
  else
    fetch_array_size=INT_OPTION(ARRAYSIZE);
  for (x=0;x<(int)ncols;x++)
  {
    if ((pcols[x].indicator=(short*)malloc(fetch_array_size*sizeof(short)))==NULL)
    {
      WPrintf("Can't perform malloc!\n");
      return 1;
    }
    if (pcols[x].datatype==OCI_TYPECODE_CLOB || pcols[x].datatype==OCI_TYPECODE_BLOB || pcols[x].datatype==OCI_TYPECODE_BFILE)
    {
      if ((pcols[x].value.lob_value=(OCILobLocator**)malloc(fetch_array_size*sizeof(OCILobLocator*)))==NULL)
      {
        WPrintf("Can't perform malloc!\n");
        return 1;
      }
      for (y=0;y<fetch_array_size;y++)
      {
        len=pcols[x].col_size;
        defhp=NULL;
        if (ChkOCIErr(OCIDescriptorAlloc(g_oci->envhp,
                                         &pcols[x].value.lob_value[y],
                                         OCI_DTYPE_LOB,
                                         0,
                                         NULL)))
        {
          return -6;
        }
      }
      if (lob_buffer==NULL)
      {
        lob_buffer=malloc(INT_OPTION(LONG_SIZE)+1);
        if (lob_buffer==NULL)
        {
          WPrintf("Can't perform malloc!\n");
          return -7;
        }
      }
      len=pcols[x].col_size;
      defhp=NULL;
      if (ChkOCIErr(OCIDefineByPos(g_oci->stmtp,
                                   &defhp,
                                   g_oci->errhp,
                                   x+1,
                                   pcols[x].value.lob_value,
                                   sizeof(OCILobLocator*),
                                   pcols[x].datatype,
                                   pcols[x].indicator,
                                   NULL,
                                   NULL,
                                   OCI_DEFAULT)))
      {
        break;
      }
    }
    else
    {
      if ((pcols[x].value.str_value=(char*)malloc(fetch_array_size*(pcols[x].col_size)))==NULL)
      {
        WPrintf("Can't perform malloc!\n");
        return 1;
      }
 	    char_ptr=pcols[x].value.str_value;
      len=pcols[x].col_size;
      defhp=NULL;
      if (ChkOCIErr(OCIDefineByPos(g_oci->stmtp,
                                   &defhp,
                                   g_oci->errhp,
                                   x+1,
                                   char_ptr,
                                   len,
                                   SQLT_STR,
                                   pcols[x].indicator,
                                   NULL,
                                   NULL,
                                   OCI_DEFAULT)))
      {
        break;
      }
    }
  }
  *p_ncols=ncols;
  *p_cols=pcols;
  //PrintHeaders(text,cols,ncols);

  return 0;
}

int OutputColumn(TextWidget *text,int y,ColumnDesc *cols,char *buffer)
{
  DWORD bytes_written;

  if (INT_OPTION(SCREEN))
  {
    if (y)
      WPrintf(" %-*s",MAX(cols[y].col_size-1,strlen(cols[y].col_name)),buffer);
    else
      WPrintf("\r%-*s",MAX(cols[y].col_size-1,strlen(cols[y].col_name)),buffer);
  }
  else if (text->spoolfile!=NULL)
  {
    if (y)
      FPrintf(" %-*s",MAX(cols[y].col_size-1,strlen(cols[y].col_name)),buffer);
    else
      FPrintf("\r%-*s",MAX(cols[y].col_size-1,strlen(cols[y].col_name)),buffer);
  }
  if (INT_OPTION(LISTVIEW))
  {
    if (WriteFile(f,buffer,cols[y].col_size,&bytes_written,NULL)==0)
    {
      DisplayLastError(NULL);
      return -1;
    }
  }
  return 0;
}

int FetchData(TextWidget *text,ColumnDesc *cols,int ncols,char *action)
{
  ub4 qrows,amount,offset;
  int result,nrows,x,y;
  sword ret;

  query_finished=0;
  total_qrows=0;
  while (!query_finished)
  {
    result=OCIStmtFetch(g_oci->stmtp,
                        g_oci->errhp,
                        fetch_array_size,
                        OCI_FETCH_NEXT,
                        OCI_DEFAULT);
    if (result!=OCI_SUCCESS && result!=OCI_NO_DATA && result!=OCI_SUCCESS_WITH_INFO)
    {
      if (ChkOCIErrI(result,1406))
        break;
    }
    if (ChkOCIErr(OCIAttrGet(g_oci->stmtp,
                             OCI_HTYPE_STMT, 
                             &qrows,
                             NULL,
                             OCI_ATTR_ROW_COUNT, 
                             g_oci->errhp)))
    {
      break;
    } 


    nrows = qrows - total_qrows;
    if (total_qrows==0 && qrows)
      PrintHeaders(text,cols,ncols);
    if (nrows!=fetch_array_size)
      query_finished=1;
    if (action[0]!='\0' && INT_OPTION(FEEDBACK))
    {
      HANDLE tmp=text->spoolfile;

      text->spoolfile=NULL;
      WPrintfOnly(action,total_qrows,total_qrows==1?"":"s");
      text->spoolfile=tmp;
    }


    for (x=0;x<nrows;x++)
    {
      text->noredraw=1;
      for (y=0;y<ncols;y++)
      {
        if (cols[y].datatype==OCI_TYPECODE_CLOB || cols[y].datatype==OCI_TYPECODE_BLOB || cols[y].datatype==OCI_TYPECODE_BFILE)
        {
          if (cols[y].indicator[x]==-1)
          {
            memset(lob_buffer,0,cols[y].col_size);
            if (OutputColumn(text,y,cols,lob_buffer))
              return total_qrows;
          }
          else
          {
            if (cols[y].datatype==OCI_TYPECODE_BFILE)
            {
              if (ChkOCIErr(OCILobFileOpen(g_oci->svchp,
                                       g_oci->errhp,
                                       cols[y].value.lob_value[x],
                                       OCI_FILE_READONLY)))
              {
                return -8;
              }
            }
            offset=1;
            do
            {
              amount=0;
              ret=OCILobRead(g_oci->svchp,
                             g_oci->errhp,
                             cols[y].value.lob_value[x],
                             &amount,
                             offset,
                             lob_buffer,
                             INT_OPTION(LONG_SIZE),
                             0,
                             NULL,
                             NULL,
                             SQLCS_IMPLICIT);
              if (ret!=OCI_SUCCESS && ret!=OCI_NEED_DATA)
              {
                if (ChkOCIErr(ret))
                {
                  return -9;
                }
              }
              if (offset==1)
              {
                lob_buffer[amount]='\0';
                if (OutputColumn(text,y,cols,lob_buffer))
                  return total_qrows;
              }
              offset+=amount;
            } while (ret==OCI_NEED_DATA);
            if (cols[y].datatype==OCI_TYPECODE_BFILE)
            {
              if (ChkOCIErr(OCILobFileClose(g_oci->svchp,
                                        g_oci->errhp,
                                        cols[y].value.lob_value[x])))
              {
                return -10;
              }
            }
          }
        }
        else
        {
          if (cols[y].indicator[x]==-1)
            memset(&cols[y].value.str_value[x*cols[y].col_size],0,cols[y].col_size);
          if (OutputColumn(text,y,cols,&cols[y].value.str_value[x*cols[y].col_size]))
            return total_qrows;
        }
      }
      text->noredraw=0;
      if (INT_OPTION(SCREEN))
        WPrintfNL();
      else if (text->spoolfile!=NULL)
        FPrintf("\n");
    }
    total_qrows = qrows;
  }
  return total_qrows;
}

int FinishSource(ColumnDesc *cols,int ncols)
{
  int x,y;

  if (ora_connected && g_oci->stmtp)
    OCIHandleFree(g_oci->stmtp,OCI_HTYPE_STMT);
  g_oci->stmtp=NULL;
  for (x=0;x<ncols;x++)
  {
    if (cols[x].datatype==OCI_TYPECODE_CLOB || cols[x].datatype==OCI_TYPECODE_BLOB || cols[x].datatype==OCI_TYPECODE_BFILE)
    {
      if (cols[x].value.lob_value)
      {
        for (y=0;y<fetch_array_size;y++)
        {
          OCIDescriptorFree(cols[x].value.lob_value[y],OCI_DTYPE_LOB);
        }
        free(cols[x].value.lob_value);
        cols[x].value.lob_value=NULL;
      }
    }
    else
    {
      if (cols[x].value.str_value)
      {
        free(cols[x].value.str_value);
        cols[x].value.str_value=NULL;
      }
    }
    free(cols[x].indicator);
  }
  if (lob_buffer!=NULL)
  {
    free(lob_buffer);
    lob_buffer=NULL;
  }
  return -1;
}


void FreeCols(ColumnDesc **cols,int *ncols)
{
  if (*cols)
  {
    free(*cols);
    *cols=NULL;
    *ncols=0;
  }
}