#include "jbsql.h"

void StrCpy(Keyword *c1,Keyword *c2,int len)
{
  c1->keyword=strdup(c2->keyword);
  c1->value=strdup(c2->value);
}

void StrFree(Keyword *c1)
{
  free(c1->keyword);
  free(c1->value);
  free(c1);
}

unsigned StrHash(char **ptr,int len,unsigned top)
{
  int x=0;
  unsigned total=0,g=0;
  
  while (x<(int)strlen(*ptr))
  {
    total=(total<<4)+(*ptr)[x++];
    if ((g=(total&0xf0000000))!=0)
    {
      total=total^(g>>24);
      total=total^g;
    }
  }
  total^=(total/top);
  return total%top;
}

int StrCmp(char  **k1,char **k2,int len)
{
  return strcmp(*k1,*k2);
}

HashTable CreateVariableHashTable(void)
{
  HashTable ht;

  ht=HTCreate(101,sizeof(char*),sizeof(Keyword),(HTHashProc)StrHash,(HTCopyProc)StrCpy,(HTFreeProc)StrFree,(HTCompareProc)StrCmp);

  return ht;
}

void DefineVariable(char *str)
{
  char *variable,*value=NULL;
  int x=0;
  Keyword k;
  
  variable=(char*)malloc(strlen(str)+1);
  while (*str && *str!='=')
  {
    variable[x]=*str;
    variable[x+1]='\0';
    x++;
    str++;
  }
  if (!*str)
  {
    if (!HTGetRecord(vht,&variable,&k))
      WPrintf("Variable %s is UNDEFINED\n",variable);
    else
    {
      WPrintf("Variable %s is defined as: \"%s\"\n",variable,k.value);
      free(k.keyword);
      free(k.value);
    }
    free(variable);
    return;
  }
  value=str+1;
  k.keyword=strdup(variable);
  k.value=strdup(value);
  Trim(k.keyword);
  Trim(k.value);
  HTDeleteRecord(vht,&k.keyword);
  HTAddRecord(vht,&k);
  free(k.keyword);
  free(k.value);
  free(variable);
}

void AcceptVariable(TextWidget *text,char **tokens,int ntokens)
{
  int x,hide=0,ok=0;
  char *variable=NULL,*format=NULL,*default_value=NULL,*prompt=NULL,value[1000];
  VariableType vt=V_CHAR;
  Keyword k;

  if (ntokens==1)
  {
    WPrintf("Usage: ACC[EPT] variable [NUM[BER]|CHAR|DATE] [FOR[MAT] format] [DEF[AULT] default] [PROMPT text|NOPR[OMPT]] [HIDE]\n");
    return;
  }
  for (x=1;x<ntokens;x++)
  {
    if (x==1)
    {
      variable=strdup(tokens[x]);
    }
    else if (!_stricmp(tokens[x],"num") || !_stricmp(tokens[x],"numb") || !_stricmp(tokens[x],"numbe") || !_stricmp(tokens[x],"number"))
    {
      vt=V_NUMBER;
    }
    else if (!_stricmp(tokens[x],"char"))
    {
      vt=V_CHAR;
    }
    else if (!_stricmp(tokens[x],"date"))
    {
      vt=V_DATE;
    }
    else if (!_stricmp(tokens[x],"for") || !_stricmp(tokens[x],"form") || !_stricmp(tokens[x],"forma") || !_stricmp(tokens[x],"format"))
    {
      x++;
      if (x<ntokens)
        format=strdup(tokens[x]);
      else
      {
        free(variable);
        if (prompt)
          free(prompt);
        if (default_value)
          free(default_value);
        WPrintf("Format mask expected after FORMAT\n");
        return;
      }
    }
    else if (!_stricmp(tokens[x],"def") || !_stricmp(tokens[x],"defa") || !_stricmp(tokens[x],"defau") || !_stricmp(tokens[x],"defaul") || !_stricmp(tokens[x],"default"))
    {
      x++;
      if (x<ntokens)
        default_value=strdup(tokens[x]);
      else
      {
        free(variable);
        if (format)
          free(format);
        if (prompt)
          free(prompt);
        WPrintf("Value expected after DEFAULT\n");
        return;
      }
    }
    else if (!_stricmp(tokens[x],"prompt"))
    {
      x++;
      if (x<ntokens)
        prompt=strdup(tokens[x]);
      else
      {
        free(variable);
        if (format)
          free(format);
        if (default_value)
          free(default_value);
        WPrintf("Value expected after PROMPT\n");
        return;
      }
    }
    else if (!_stricmp(tokens[x],"nopr") || !_stricmp(tokens[x],"nopro") || !_stricmp(tokens[x],"noprom") || !_stricmp(tokens[x],"nopromp") || !_stricmp(tokens[x],"noprompt"))
    {
      if (prompt)
        free(prompt);
      prompt=NULL;
    }
    else if (!_stricmp(tokens[x],"hide"))
    {
      hide=1;
    }
    else
    {
      WPrintf("Usage: ACC[EPT] variable [NUM[BER]|CHAR|DATE] [FOR[MAT] format] [DEF[AULT] default] [PROMPT text|NOPR[OMPT]] [HIDE]\n");
      free(variable);
      if (format)
        free(format);
      if (prompt)
        free(prompt);
      if (default_value)
        free(default_value);
      return;
    }
  }
  while (!ok)
  {
    if (prompt)
    {
      WPrintf("%s",prompt);
    }
    WScanf(text,value,hide);
    if (value[0]=='\0' && default_value)
      strcpy(value,default_value);
    if (vt==V_CHAR || CheckVariableFormat(vt,format,value)==0)
      ok=1;
  }
  HTDeleteRecord(vht,&variable);
  k.keyword=strdup(variable);
  k.value=strdup(value);
  HTAddRecord(vht,&k);
  free(k.keyword);
  free(k.value);
  free(variable);
  if (format)
    free(format);
  if (prompt)
    free(prompt);
  if (default_value)
    free(default_value);
}